﻿using BasicAuthLogon.Models;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using static BasicAuthLogon.AccessTokenDecoder;

namespace BasicAuthLogon
{
   internal class Program
   {
      static void Main(string[] args)
      {
         var app = new App();
         TokenInfo tokenInfo = app.GetToken(out HttpStatusCode code);
         if (tokenInfo == null)
         {
            Console.WriteLine($"Unable to retrieve token. Code: {code}");
            return;
         }

         Console.WriteLine("Token Info:");
         Console.WriteLine("===========");
         Console.WriteLine("token_type   : " + tokenInfo.token_type);
         Console.WriteLine("access_token : " + tokenInfo.access_token);
         Console.WriteLine("id_token     : " + tokenInfo.id_token);
         Console.WriteLine("refresh_token: " + tokenInfo.refresh_token);
         Console.WriteLine("scope        : " + tokenInfo.scope);
         Console.WriteLine("expires_in   : " + tokenInfo.expires_in);

         // Extract the data center hosting this user (encoded as a claim within the access token)
         // 
         string dataCenterURI = ExtractDataCenterURI(tokenInfo.access_token);

         // Demonstrate calling the Storage APIs
         try
         {
            SpacesCollection spaces = RetrieveSpaces(dataCenterURI, tokenInfo).Result;
            if (spaces.Spaces.Count > 0)
            {
               Console.WriteLine("\nAvailable Spaces:");
               foreach (Space space in spaces.Spaces)
                  Console.WriteLine($"  {space.DisplayName}");
            }
         }
         catch (Exception ex) 
         {
            Console.WriteLine($"Exception occurred while retrieving available spaces. Ex: {ex.Message}");
         }
      }

      static string ExtractDataCenterURI(string accessToken)
      {
         DecodedToken decodedToken = AccessTokenDecoder.Decode(accessToken);

         if (decodedToken.Payload.ContainsKey(DataCenterClaim))
            return decodedToken.Payload[AccessTokenDecoder.DataCenterClaim] as string;
         throw new Exception("DataCenterURI missing");
      }

      static async Task<SpacesCollection> RetrieveSpaces(string dataCenterUri, TokenInfo tokenInfo)
      {
         string GetSpaces_Url = "api/librarian/spaces?includeHidden={0}";
         bool includeHidden = false;

         HttpClient httpClient = new HttpClient();
         httpClient.DefaultRequestHeaders.Clear();
         httpClient.BaseAddress = new Uri(dataCenterUri);
         httpClient.DefaultRequestHeaders.Add("Authorization", $"Bearer {tokenInfo.access_token}");

         try
         {
            string requestUrl = string.Format(GetSpaces_Url, includeHidden);
            HttpResponseMessage msg = await httpClient.GetAsync(requestUrl);

            if (msg.IsSuccessStatusCode)
            {
               string json = await msg.Content.ReadAsStringAsync();
               return JsonConvert.DeserializeObject<SpacesCollection>(json) ?? new SpacesCollection();
            }
            else
               throw new Exception(msg.ReasonPhrase);
         }
         catch (Exception)
         {
            throw;
         }
      }
   }

   class TokenInfo
   {
      public string access_token { get; set; }
      public string refresh_token { get; set; }
      public string id_token { get; set; }
      public string scope { get; set; }
      public int expires_in { get; set; } = 0;
      public string token_type { get; set; }
   }

   class App
   {
      public static string ClaimsIssuer { get; set; } = string.Empty;

      public TokenInfo GetToken(out HttpStatusCode code)
      {
         TokenInfo token = null;

         Console.WriteLine("Please enter information about an Native application created\nwith grant_type:\"password refresh_token\"\n");

         string clientId;     // BarTenderCloud app client-id
         string clientSecret; // BarTenderCloud app client-secret
         string username;
         string password;

         do
         {
            // 
            Console.Write("BarTenderCloud Cluster (e.g. https://am1.bartendercloud.com) : ");
            string input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
               try
               {
                  RetrieveAuthenticationConfiguration(input);
                  break;
               }
               catch (Exception ex)
               {
                  Console.WriteLine($"Unable to retrieve OIDC document. Ex={ex.Message}");
                  continue;
               }
            }
         } while (true);

         do
         {
            Console.Write("Application Client Id: ");
            string input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
               clientId = input;
               break;
            }
         } while (true);

         do
         {
            Console.Write("Application Client Secret: ");
            string input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
               clientSecret = input;
               break;
            }
         } while (true);

         do
         {
            Console.Write("Username: ");
            string input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
               username = input;
               break;
            }
         } while (true);

         do
         {
            Console.Write("Password: ");
            string input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
               password = input;
               break;
            }
         } while (true);

         HttpClient client = new HttpClient();
         Uri uri = new Uri(ClaimsIssuer);
         client.DefaultRequestHeaders.Accept.Add(new System.Net.Http.Headers.MediaTypeWithQualityHeaderValue("application/x-www-form-urlencoded"));
         client.BaseAddress = uri;
         var contentBodyList = new List<KeyValuePair<string, string>>();
         contentBodyList.Add(new KeyValuePair<string, string>("grant_type", "password"));
         contentBodyList.Add(new KeyValuePair<string, string>("username", username));
         contentBodyList.Add(new KeyValuePair<string, string>("password", password));
         contentBodyList.Add(new KeyValuePair<string, string>("client_id", clientId));
         contentBodyList.Add(new KeyValuePair<string, string>("client_string", clientSecret));
         contentBodyList.Add(new KeyValuePair<string, string>("audience", "https://BarTenderCloudServiceApi"));
         contentBodyList.Add(new KeyValuePair<string, string>("scope", "openid profile user:query offline_access"));

         var request = new HttpRequestMessage(HttpMethod.Post, "oauth/token") { Content = new FormUrlEncodedContent(contentBodyList) };
         HttpResponseMessage msg = client.SendAsync(request).Result;

         code = msg.StatusCode;
         if (msg.IsSuccessStatusCode)
         {
            var resultPayload = msg.Content.ReadAsStringAsync().Result;
            token = JsonConvert.DeserializeObject<TokenInfo>(resultPayload);
         }

         return token;
      }

      static void RetrieveAuthenticationConfiguration(string barTenderCloudCluster)
      {
         HttpClient httpClient = new HttpClient();
         string url = barTenderCloudCluster + (!barTenderCloudCluster.EndsWith('/') ? "/" : "") + ".well-known/openid-configuration";
         HttpResponseMessage msg = httpClient.GetAsync(url).Result;
         if (!msg.IsSuccessStatusCode)
            throw new Exception("Failure to retrieve OIDC discovery document");

         string result = msg.Content.ReadAsStringAsync().Result;
         IDictionary<string, object> dictionary = JsonConvert.DeserializeObject<IDictionary<string, object>>(result)!;
         var i = (string)dictionary["issuer"];
         ClaimsIssuer = i.EndsWith('/') ? i.TrimEnd(new char[] { '/' }) : i;
      }
   }

   /// <summary>
   /// Decode an access token
   /// </summary>
   internal static class AccessTokenDecoder
   {
      public static string TenantIDClaim = "https://BarTenderCloud.com/TenantID";
      public static string UserIDClaim = "https://BarTenderCloud.com/UserID";
      public static string DataCenterClaim = "https://BarTenderCloud.com/DataCenterURI";

      public class DecodedToken
      {
         public Dictionary<string, string> Header { get; set; } = new();
         public Dictionary<string, object> Payload { get; set; } = new();
         public string Signature { get; set; } = String.Empty;
      }

      /// <summary>
      /// Decode an access token (NOTE: Does not validate the token signature)
      /// </summary>
      /// <param name="encodedToken">access token to validate</param>
      /// <returns>DecodedToken</returns>
      /// <throws>Exception on invalid token</throws>
      public static DecodedToken Decode(string encodedToken)
      {
         DecodedToken decodedToken = new();

         try
         {
            // Split the supplied access token into it's three parts
            string[] parts = encodedToken.Split('.');
            if (parts.Length != 3)
               throw new Exception("Invalid Access Token. Incorrect number of parts");

            string header = DecodeBase64(parts[0]);
            string payload = DecodeBase64(parts[1]);
            string signature = DecodeBase64(parts[2]);

            var headerDictionary = JsonConvert.DeserializeObject<Dictionary<string, string>>(header);
            if (headerDictionary == null)
               throw new Exception("Invalid Access Token. Header empty");

            var payloadDictionary = JsonConvert.DeserializeObject<Dictionary<string, object>>(payload);
            if (payloadDictionary == null)
               throw new Exception("Invalid Access Token. Payload empty");

            // Validate the token expiration
            if (!payloadDictionary!.ContainsKey("exp"))
               throw new Exception("Invalid Access Token. Expiration missing");

            decodedToken.Signature = signature;
            decodedToken.Header = headerDictionary;
            decodedToken.Payload = payloadDictionary;
         }
         catch
         {
            throw;
         }

         return decodedToken;
      }

      // Decode a Base64 string to a string
      private static string DecodeBase64(string value)
      {
         if (string.IsNullOrEmpty(value))
            return string.Empty;
         value = value.Replace(' ', '+').Replace('-', '+').Replace('_', '/').PadRight(4 * ((value.Length + 3) / 4), '=');

         var valueBytes = System.Convert.FromBase64String(value);
         return System.Text.Encoding.UTF8.GetString(valueBytes);
      }
   }
}
